#include <mpi.h>
#include "neighbours.h"

void main() {

 const int ndims=2, num_neigh=4, reorder=1;
 int dims[ndims];
 int neigh_rank[num_neigh], own_coords[ndims], i, j, it;
 int periods[ndims];
 double u[102][102], f[102][102];
 MPI_Comm comm, comm_cart;
 
 comm = MPI_COMM_WORLD;
 
 //Set process grid size and periodicity
 MPI_Dims_create(comm, ndims, &dims[0]);
 periods[0] = 1;
 periods[1] = 1;
 
 //Create a grid structure in WORLD group and inquire about own position
 MPI_Cart_create(comm, ndims, dims, periods, reorder, &comm_cart);
 MPI_Cart_get(comm_cart, ndims, dims, periods, &own_coords[0]);
 i = own_coords[0];
 j = own_coords[1];
 
 //Look up the ranks for the neighbors. Own process coordinates are (i,j).
 //Neighbors are (i-1,j), (i+1,j), (i,j-1), (i,j+1) modulo (dims(1),dims(2))
 MPI_Cart_shift(comm_cart, 0, 1, &neigh_rank[0], &neigh_rank[1]);
 MPI_Cart_shift(comm_cart, 1, 1, &neigh_rank[2], &neigh_rank[3]);
 
 MPI_Aint lb, sizeofreal;
 MPI_Type_get_extent(MPI_DOUBLE, &lb, &sizeofreal);
 
 MPI_Datatype type_vec;
 MPI_Type_vector(100, 1, 102, MPI_DOUBLE, &type_vec);
 MPI_Type_commit(&type_vec);
 
 int sndcounts[num_neigh], sndtypes[num_neigh];
 sndtypes[0] = type_vec;
 sndtypes[1] = type_vec;
 sndtypes[2] = MPI_DOUBLE;
 sndtypes[3] = MPI_DOUBLE;
 sndcounts[0] = 1;
 sndcounts[1] = 1;
 sndcounts[2] = 100;
 sndcounts[3] = 100;
 
 int rcvcounts[num_neigh], rcvtypes[num_neigh];
 rcvtypes[0] = type_vec;
 rcvtypes[1] = type_vec;
 rcvtypes[2] = MPI_DOUBLE;
 rcvtypes[3] = MPI_DOUBLE;
 rcvcounts[0] = 1;
 rcvcounts[1] = 1;
 rcvcounts[2] = 100;
 rcvcounts[3] = 100;
 
 MPI_Aint sdispls[num_neigh];
 sdispls[0] = (  1 +   1*102) * sizeofreal; // first element of u(    1,1:100)
 sdispls[1] = (100 +   1*102) * sizeofreal; // first element of u(  100,1:100)
 sdispls[2] = (  1 +   1*102) * sizeofreal; // first element of u(1:100,    1)
 sdispls[3] = (  1 + 100*102) * sizeofreal; // first element of u(1:100,  100)
 
 MPI_Aint rdispls[num_neigh];
 rdispls[0] = (  0 +   1*102) * sizeofreal; // first element of u(    0,1:100)
 rdispls[1] = (101 +   1*102) * sizeofreal; // first element of u(  101,1:100)
 rdispls[2] = (  1 +   0*102) * sizeofreal; // first element of u(1:100,    0)
 rdispls[3] = (  1 + 101*102) * sizeofreal; // first element of u(1:100,  101)
 
 // initialize the grid functions
 init (&u[0][0], &f[0][0]);

 for (int it=0;it<100;it++)
 {
  relax(&u[0][0], &f[0][0]);
  
  // the following communication needs to be done in each iteration
  MPI_Neighbor_alltoallw(u, sndcounts, sdispls, sndtypes,
                         u, rcvcounts, rdispls, rcvtypes, comm_cart);
 }
 
 output(&u[0][0]);
 
 // the following finalisation needs to be done only once
 // after the last communication call
 MPI_Type_free(&type_vec);
 MPI_Comm_free(&comm_cart);
 MPI_Finalize();
};
